/* This file is part of tagr.
   Copyright (C) 2000, 2005,  Max Bouglacoff, Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <pwd.h>
#include <grp.h>
#include <tagr.h>
#include <report.h>

#ifndef RETSIGTYPE
# define RETSIGTYPE void
#endif

static RETSIGTYPE sig_quit (int);
static RETSIGTYPE sig_fatal (int);
static RETSIGTYPE sig_hup (int);
static RETSIGTYPE sig_child (int);

char pidfile[] = TAGR_PIDFILE;
char i_recv_buffer[reply_size (MAXADDR)];
char *recv_buffer = i_recv_buffer;
char *configfile = TAGR_CONFIGFILE;
char *html_template = TAGR_TEMPLATE;
char *user;
char *basedir;
int port;
int sockfd;
int foreground = 0;
int single_process = 0;

/* Change to the given uid/gid. Clear the supplementary group list.
   On success returns 0.
   On failure returns 1 (or exits, depending on topt settings. See
   anubis_error) */
static int
change_privs (uid_t uid, gid_t gid)
{
  int rc = 0;
  gid_t emptygidset[1];

  /* Reset group permissions */
  emptygidset[0] = gid ? gid : getegid ();
  if (geteuid () == 0 && setgroups (1, emptygidset))
    {
      logmsg (L_ERR, "setgroups(1, %lu) failed: %s",
	      (u_long) emptygidset[0], strerror (errno));
      rc = 1;
    }

  /* Switch to the user's gid. On some OSes the effective gid must
     be reset first */

#if defined(HAVE_SETEGID)
  if ((rc = setegid (gid)) < 0)
    logmsg (L_ERR, "setegid(%lu) failed: %s", (u_long) gid, strerror (errno));
#elif defined(HAVE_SETREGID)
  if ((rc = setregid (gid, gid)) < 0)
    logmsg (L_ERR, "setregid(%lu,%lu) failed: %s",
	    (u_long) gid, (u_long) gid, strerror (errno));
#elif defined(HAVE_SETRESGID)
  if ((rc = setresgid (gid, gid, gid)) < 0)
    logmsg (L_ERR, "setresgid(%lu,%lu,%lu) failed: %s",
	    (u_long) gid, (u_long) gid, (u_long) gid, strerror (errno));
#endif

  if (rc == 0 && gid != 0)
    {
      if ((rc = setgid (gid)) < 0 && getegid () != gid)
	logmsg (L_ERR, "setgid(%lu) failed: %s",
		(u_long) gid, strerror (errno));
      if (rc == 0 && getegid () != gid)
	{
	  logmsg (L_ERR, "cannot set effective gid to %lu: %s",
		  (u_long) gid, strerror (errno));
	  rc = 1;
	}
    }

  /* now reset uid */
  if (rc == 0 && uid != 0)
    {
      uid_t euid;

      if (setuid (uid)
	  || geteuid () != uid
	  || (getuid () != uid && (geteuid () == 0 || getuid () == 0)))
	{

#if defined(HAVE_SETREUID)
	  if (geteuid () != uid)
	    {
	      if (setreuid (uid, -1) < 0)
		{
		  logmsg (L_ERR,
			  "setreuid(%lu,-1) failed: %s",
			  (u_long) uid, strerror (errno));
		  rc = 1;
		}
	      if (setuid (uid) < 0)
		{
		  logmsg (L_ERR,
			  "second setuid(%lu) failed: %s",
			  (u_long) uid, strerror (errno));
		  rc = 1;
		}
	    }
	  else
#endif
	    {
	      logmsg (L_ERR, "setuid(%lu) failed: %s",
		      (u_long) uid, strerror (errno));
	      rc = 1;
	    }
	}


      euid = geteuid ();
      if (uid != 0 && setuid (0) == 0)
	{
	  logmsg (L_ERR, "seteuid(0) succeeded when it should not");
	  rc = 1;
	}
      else if (uid != euid && setuid (euid) == 0)
	{
	  logmsg (L_ERR, "cannot drop non-root setuid privileges");
	  rc = 1;
	}
    }
  return rc;
}

void
change_user ()
{
  struct passwd *pwd;

  if (user == NULL)
    return;

  pwd = getpwnam (user);
  if (pwd)
    {
      if (change_privs (pwd->pw_uid, pwd->pw_gid))
	exit (1);

      chdir (pwd->pw_dir);
    }
}


int
main (int argc, char **argv)
{
  int c, rc;
  struct sockaddr salocal, saremote;
  struct sockaddr_in *sin;
  fd_set read_fds;
  int salen;
  char *check_mode = 0;
  char *opt_user = NULL;
  char *opt_html_template = NULL;
  int rebuild_option = 0;
  
  init_syslog (argv[0]);

  while ((c = getopt (argc, argv, "bc:fst:T:u:")) != EOF)
    {
      switch (c)
	{
	case 'f':
	  foreground++;
	  break;
	  
	case 'b':
	  rebuild_option = 1;
	  break;
	  
	case 'c':
	  configfile = optarg;
	  break;

	case 's':
	  single_process++;
	  break;
	  
	case 't':
	  opt_html_template = optarg;
	  break;
	  
	case 'u':
	  opt_user = optarg;
	  break;
	  
	case 'T':
	  check_mode = optarg;
	  break;
	  
	default:
	  logmsg (L_ERR, "invalid switch: -%c", c);
	  exit (1);
	}
    }


  if (check_mode)
    for (; *check_mode; check_mode++)
      switch (*check_mode)
	{
	case 'c':
	  exit (readconfig ());

	case 't':
	  exit (check_template ());

	default:
	  fprintf (stderr, "Unknown check letter: %c", *check_mode);
	  exit (1);
	}

  if (readconfig ())
    exit (2);

  if (opt_user)
    user = opt_user;
  if (opt_html_template)
    html_template = opt_html_template;

  if (rebuild_option)
    {
      rebuild ();
      return 0;
    }
  
  sockfd = socket (AF_INET, SOCK_DGRAM, 0);
  if (sockfd < 0)
    {
      logmsg (L_ERR, "socket: %s", strerror (errno));
      exit (1);
    }
  sin = (struct sockaddr_in *) &salocal;
  memset (sin, 0, sizeof (salocal));
  sin->sin_family = AF_INET;
  sin->sin_addr.s_addr = INADDR_ANY;
  sin->sin_port = htons (port);

  if (bind (sockfd, &salocal, sizeof (*sin)) < 0)
    {
      logmsg (L_ERR, "bind: %s", strerror (errno));
      exit (1);
    }

  signal (SIGHUP, sig_hup);
  signal (SIGUSR1, SIG_IGN);
  signal (SIGUSR2, SIG_IGN);
  signal (SIGQUIT, sig_quit);
  signal (SIGTERM, sig_quit);
  /*signal(SIGTRAP, sig_fatal); */
  signal (SIGIOT, sig_fatal);
  signal (SIGCHLD, sig_child);
  signal (SIGPIPE, SIG_IGN);

  if (!foreground)
    {
      int i;
      pid_t pid;

      pid = fork ();
      if (pid < 0)
	die (1, "cannot fork: %s", strerror (errno));
      if (pid > 0)
	{
	  /* Parent branch */
	  FILE *fp;
	  /* write pid file */
	  if ((fp = fopen (pidfile, "w")) != NULL)
	    {
	      fprintf (fp, "%d\n", pid);
	      fclose (fp);
	    }
	  else
	    {
	      logmsg (L_ERR, "can't write pid file %s: %s",
		      pidfile, strerror (errno));
	    }
	  exit (0);
	}
      /* Child branch */
      chdir ("/");
      for (i = 0; i < 3; i++)
	close (i);
      setsid ();
      use_stderr = 0;
    }

  //change_user();

  sin = (struct sockaddr_in *) &saremote;
  for (;;)
    {
      FD_ZERO (&read_fds);
      FD_SET (sockfd, &read_fds);

      rc = select (sockfd + 1, &read_fds, NULL, NULL, NULL);
      if (rc < 0)
	{
	  if (errno != EINTR)
	    logmsg (L_ERR, "select: %s", strerror (errno));
	  continue;
	}
      if (rc == 1)
	{
	  rc = recvfrom (sockfd, (char *) recv_buffer,
			 sizeof (i_recv_buffer), 0, &saremote, &salen);
	  if (rc > 0)
	    decode_buffer ();
	  else if (rc < 0 && errno != EINTR)
	    logmsg (L_ERR, "recvfrom: %s", strerror (errno));
	}
    }

}

void
decode_buffer ()
{
  int i;
  Stat_reply *reply;
  Stat *sp;
  int child = 0;

  reply = (Stat_reply *) recv_buffer;
  reply->n_addr = ntohl (reply->n_addr);

  if (reply->n_addr > MAXADDR)
    {
      logmsg (L_NOTICE, "got invalid packet: n_addr = %d", reply->n_addr);
      return;
    }
  reply->timestamp = ntohl (reply->timestamp);
  sp = reply->stat;

  if (!single_process)
    {
      pid_t pid = fork ();
      if (pid > 0)
	return;
      else if (pid < 0)
	logmsg (L_ERR, "can't fork: %s", strerror (errno));
      else
	{
	  signal (SIGHUP, SIG_IGN);
	  signal (SIGCHLD, SIG_IGN);
	  child = 1;
	}
    }

  for (i = 0; i < reply->n_addr; i++, sp++)
    {
      sp->in = ntohl (sp->in);
      sp->out = ntohl (sp->out);
      report (sp);
    }

  if (child)
    exit (0);
}

int
get_port (char *str)
{
  int pn;

  if (isdigit (str[0]))
    pn = htons (atoi (str));
  else
    {
      struct servent *s = getservbyname (str, "udp");
      if (s)
	pn = s->s_port;
      else
	{
	  logmsg (L_ERR, "no such service: %s", str);
	  return 0;
	}
    }
  return pn;
}

RETSIGTYPE
sig_quit (int sig)
{
  logmsg (L_INFO, "exiting on signal %d", sig);
  unlink (pidfile);
  exit (0);
}

RETSIGTYPE
sig_fatal (int sig)
{
  logmsg (L_ERR, "FATAL: exiting on signal %d", sig);
  unlink (pidfile);
  exit (1);
}

RETSIGTYPE
sig_hup (int sig)
{
  signal (sig, sig_hup);
}

RETSIGTYPE
sig_child (int sig)
{
  int status;

  waitpid ((pid_t) - 1, &status, WNOHANG);
  signal (sig, sig_child);
}
